# frozen_string_literal: true

module DrgDSL
  class ParenCleaner
    include Visitor

    # @param expression [String]
    # @return [String] pretty printed expression without any redundant
    #   parentheses.
    def self.remove_redundant_parens(expression)
      ast = Parser.parse(expression)
      reduced_ast = clean(ast)
      reduced_ast.to_s
    end

    # @param ast [Node]
    # @return [Node] ast without redundant ParenExpressions
    def self.clean(ast, remove_outer_parens: true)
      # If the whole expression is wrapped in parentheses, we can start
      # visiting the first child.
      if ast.type == 'ParenExpression' && remove_outer_parens
        ast = ast.expression
      end

      ast.accept(new)
    end

    private

    def visit_Expression(n)
      Ast::Expression.new(n.expressions.map { |e| visit(e) })
    end

    def visit_AndExpression(n)
      Ast::AndExpression.new(n.expressions.map { |e| visit(e) })
    end

    def visit_ParenExpression(n)
      types = %w(
        ParenExpression
        DrgLink
        BasicExpression
        AndExpression
        NotExpression
        FunctionCall
        Empty
        TableCondition
      )
      return visit(n.expression) if types.include?(n.expression.type)

      if n.expression.type == 'Expression'
        return Ast::ParenExpression.new(visit(n.expression))
      end

      n
    end
  end
end
